"""Tests methods in ecephys session module"""

import json
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from slims.internal import Record

from aind_slims_api.core import SlimsClient
from aind_slims_api.models.ecephys_session import (
    SlimsStreamsResult,
    SlimsMouseSessionResult,
    SlimsStimulusEpochsResult,
)

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"


class TestEcephysSessionResults(unittest.TestCase):
    """Tests top level methods in mouse module"""

    example_client: SlimsClient
    example_fetch_mouse_response: list[Record]

    @classmethod
    def setUpClass(cls):
        """Load json files of expected responses from slims"""
        example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        cls.example_client = example_client
        with open(
            RESOURCES_DIR / "example_fetch_ecephys_session_result.json", "r"
        ) as f:
            response = [
                Record(json_entity=r, slims_api=example_client.db.slims_api)
                for r in json.load(f)
            ]
        cls.example_fetch_ecephys_session_result = response

        with open(
            RESOURCES_DIR / "example_fetch_ecephys_streams_result.json", "r"
        ) as f:
            response = [
                Record(json_entity=r, slims_api=example_client.db.slims_api)
                for r in json.load(f)
            ]
        cls.example_fetch_ecephys_streams_result = response

        with open(
            RESOURCES_DIR / "example_fetch_ecephys_stimulus_epochs_result.json", "r"
        ) as f:
            response = [
                Record(json_entity=r, slims_api=example_client.db.slims_api)
                for r in json.load(f)
            ]
        cls.example_fetch_ecephys_stimulus_epochs_result = response

    @patch("slims.slims.Slims.fetch")
    def test_fetch_ecephys_mouse_session_success(self, mock_fetch: MagicMock):
        """Test fetch_mouse_session when successful"""
        mock_fetch.return_value = self.example_fetch_ecephys_session_result
        session = self.example_client.fetch_model(
            SlimsMouseSessionResult, mouse_pk=3135
        )
        self.assertEqual(
            self.example_fetch_ecephys_session_result[0].json_entity,
            session.json_entity,
        )
        self.assertEqual(session.test_label, "Mouse Session")

    @patch("slims.slims.Slims.fetch")
    def test_fetch_ecephys_streams_success(self, mock_fetch: MagicMock):
        """Test fetch streams when successful"""
        mock_fetch.return_value = self.example_fetch_ecephys_streams_result
        streams = self.example_client.fetch_model(
            SlimsStreamsResult, mouse_session_pk=2329
        )
        self.assertEqual(
            self.example_fetch_ecephys_streams_result[0].json_entity,
            streams.json_entity,
        )
        self.assertEqual(streams.test_label, "Streams")
        self.assertEqual(streams.camera_names, ["Face camera"])

    @patch("slims.slims.Slims.fetch")
    def test_fetch_ecephys_stim_epochs_success(self, mock_fetch: MagicMock):
        """Test fetch stim epochs when successful"""
        mock_fetch.return_value = self.example_fetch_ecephys_stimulus_epochs_result
        stim_epochs = self.example_client.fetch_model(
            SlimsStimulusEpochsResult, mouse_session_pk=2329
        )
        self.assertEqual(
            self.example_fetch_ecephys_stimulus_epochs_result[0].json_entity,
            stim_epochs.json_entity,
        )
        self.assertEqual(stim_epochs.test_label, "Stimulus Epochs")
        self.assertEqual(stim_epochs.laser_name, "Coherent Red Laser")


if __name__ == "__main__":
    unittest.main()



"""Contains a model for a waterlog water restriction event stored in SLIMS."""

from datetime import datetime
from typing import ClassVar, Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel

DEFAULT_WEIGHT_FRACTION = 0.85
MIN_WEIGHT_FRACTION = 0.75
MAX_WEIGHT_FRACTION = 1.0


class SlimsWaterRestrictionEvent(SlimsBaseModel):
    """Model for a Water Restriction Event

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> from aind_slims_api import models
    >>> client = SlimsClient()
    >>> mouse = client.fetch_model(models.SlimsMouseContent, barcode="00000000")

    ### Write
    >>> water_restriction_event = client.add_model(
    ...  models.SlimsWaterRestrictionEvent(
    ...   mouse_pk=mouse.pk,
    ...   start_date=datetime(2021,1,1),
    ...   end_date=datetime(2021,1,2),
    ...   assigned_by="mochic",
    ...   target_weight_fraction=0.90,
    ...  )
    ... )

    ### Read
    >>> water_restriction_event = client.fetch_model(
    ...  models.SlimsWaterRestrictionEvent,
    ...  mouse_pk=mouse.pk,
    ... )
    >>> water_restriction_event.target_weight_fraction
    0.9
    """

    start_date: datetime = Field(
        datetime.now(),
        serialization_alias="cnvn_cf_startDate",
        validation_alias="cnvn_cf_startDate",
    )
    end_date: Optional[datetime] = Field(
        None,
        serialization_alias="cnvn_cf_endDate",
        validation_alias="cnvn_cf_endDate",
    )
    assigned_by: str = Field(
        ...,
        serialization_alias="cnvn_cf_assignedBy",
        validation_alias="cnvn_cf_assignedBy",
    )
    target_weight_fraction: float = Field(
        default=DEFAULT_WEIGHT_FRACTION,
        serialization_alias="cnvn_cf_targetWeightFraction",
        validation_alias="cnvn_cf_targetWeightFraction",
        gt=MIN_WEIGHT_FRACTION,
        lt=MAX_WEIGHT_FRACTION,
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="cnvn_pk",
        validation_alias="cnvn_pk",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="cnvn_createdOn",
        validation_alias="cnvn_createdOn",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="cnvn_fk_content",
        validation_alias="cnvn_fk_content",
    )
    cnvn_fk_contentEventType: int = 9  # pk of Water Restriction ContentEvent type

    _slims_table = "ContentEvent"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "cnvt_name": "Water Restriction",
    }



"""Utility functions and classes for working with slims models.
"""

from typing import Optional

from pydantic.fields import FieldInfo


class UnitSpec:
    """Used in type annotation metadata to specify units"""

    units: tuple[str]
    preferred_unit: Optional[str] = None

    def __init__(self, *args, preferred_unit=None):
        """Set list of acceptable units from args, and preferred_unit"""
        self.units = args
        if len(self.units) == 0:
            raise ValueError("One or more units must be specified")
        if preferred_unit is None:
            self.preferred_unit = self.units[0]


def _find_unit_spec(field: FieldInfo) -> UnitSpec | None:
    """Given a Pydantic FieldInfo, find the UnitSpec in its metadata"""
    metadata = field.metadata
    for m in metadata:
        if isinstance(m, UnitSpec):
            return m
    return None



"""Tests methods in mouse module"""

import json
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from slims.internal import Record

from aind_slims_api.core import SlimsClient
from aind_slims_api.models.instrument import SlimsInstrument

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"


class TestInstrument(unittest.TestCase):
    """Tests top level methods in mouse module"""

    example_client: SlimsClient
    example_response: Record

    @classmethod
    def setUpClass(cls):
        """Load json files of expected responses from slims"""
        cls.example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        cls.example_response = [
            Record(json_entity=r, slims_api=cls.example_client.db.slims_api)
            for r in json.loads(
                (
                    RESOURCES_DIR / "example_fetch_instrument_response.json_entity.json"
                ).read_text()
            )
        ]

    @patch("slims.slims.Slims.fetch")
    def test_fetch_content_success(
        self,
        mock_fetch: MagicMock,
    ):
        """Test fetch_instrument_content when successful and multiple are
        returned from fetch
        """
        mock_fetch.return_value = self.example_response + self.example_response
        response = self.example_client.fetch_model(
            SlimsInstrument, name="323_EPHYS1_OPTO"
        )
        self.assertEqual(response.json_entity, self.example_response[0].json_entity)


if __name__ == "__main__":
    unittest.main()



"""Tests functions in filters"""

import unittest

from slims.criteria import (
    between_inclusive,
    disjunction,
    is_na,
    is_not,
    is_not_one_of,
    is_one_of,
)

from aind_slims_api import filters
from aind_slims_api.models.unit import SlimsUnit
from aind_slims_api.models.user import SlimsUser


class TestFilters(unittest.TestCase):
    """Tests filter validation and utility functions."""

    def test_resolve_model_alias_invalid(self):
        """Tests resolve_model_alias method raises expected error with an
        invalid alias name.
        """
        with self.assertRaises(ValueError):
            filters.resolve_model_alias(SlimsUnit, "not_an_alias")

    def test__validate_field_name_failure(self):
        """Tests _validate_field_name method raises expected error with an
        invalid field name.
        """
        with self.assertRaises(ValueError):
            filters._validate_field_name(SlimsUnit, "not_an_alias")

    def test_validate_criteria_is_one_of(self):
        """Tests _validate_criteria method with an is_one_of criterion."""
        filters.validate_criteria(
            SlimsUser,
            is_one_of("username", ["LKim", "JSmith"]),
        )

    def test_validate_criteria_is_not_one_of(self):
        """Tests _validate_criteria method with an is_not_one_of
        criterion.
        """
        filters.validate_criteria(
            SlimsUser,
            is_not_one_of("username", ["LKim", "JSmith"]),
        )

    def test_validate_criteria_between_inclusive(self):
        """Tests _validate_criteria method with an between_inclusive criterion."""
        filters.validate_criteria(
            SlimsUser,
            between_inclusive("username", "LKim", "JSmith"),
        )

    def test_validate_criteria_is_na(self):
        """Tests _validate_criteria method with an is_na criterion."""
        filters.validate_criteria(
            SlimsUser,
            is_na("username"),
        )

    def test_resolve_criteria_is_na(self):
        """Tests _resolve_criteria method with an is_na criterion."""
        filters.resolve_criteria(
            SlimsUser,
            is_na("username"),
        )

    def test_validate_criteria_is_not(self):
        """Tests _validate_criteria method with an is_not criterion."""
        filters.validate_criteria(
            SlimsUser,
            is_not(is_one_of("username", ["LKim", "JSmith"])),
        )

    def test_validate_criteria_disjunction(self):
        """Tests _validate_criteria method with an is_not criterion."""
        criteria = (
            disjunction()
            .add(is_one_of("username", ["LKim"]))
            .add(is_one_of("username", ["JSmith"]))
        )
        filters.validate_criteria(
            SlimsUser,
            criteria,
        )


if __name__ == "__main__":
    unittest.main()



"""Contains a model for the behavior session content events, a method for
 fetching it and writing it.
"""

import logging
from datetime import datetime
from typing import ClassVar, Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel

logger = logging.getLogger()


class SlimsBehaviorSession(SlimsBaseModel):
    """Model for an instance of the Behavior Session ContentEvent

    Examples
    --------
    >>> from datetime import datetime
    >>> from aind_slims_api import SlimsClient
    >>> from aind_slims_api import models
    >>> client = SlimsClient()
    >>> mouse = client.fetch_model(models.SlimsMouseContent, barcode="00000000")

    ### Read
    >>> behavior_sessions = client.fetch_models(SlimsBehaviorSession,
    ...  mouse_pk=mouse.pk, sort=["date"])
    >>> curriculum_attachments = client.fetch_attachments(behavior_sessions[0])

    ### Write
    >>> trainer = client.fetch_model(models.SlimsUser, username="LKim")
    >>> instrument = client.fetch_model(models.SlimsInstrument, name="323_EPHYS1_OPTO")
    >>> added = client.add_model(
    ...  SlimsBehaviorSession(
    ...      mouse_pk=mouse.pk,
    ...      instrument_pk=instrument.pk,
    ...      trainer_pks=[trainer.pk],
    ...      notes="notes",
    ...      task_stage="stage",
    ...      task="task",
    ...      task_schema_version="0.0.1",
    ...      is_curriculum_suggestion=True,
    ...      date=datetime(2021, 1, 2),
    ...  )
    ... )

    ### Add a curriculum attachment
    Attachment content isn't available immediately.
    >>> import json
    >>> attachment_pk = client.add_attachment_content(
    ...  added,
    ...  "curriculum",
    ...  json.dumps({"curriculum_key": "curriculum_value"}),
    ... )

    Properties
    ----------
    mouse_pk : Optional[int]
        The primary key of the mouse associated with this behavior session.
    instrument_pk : Optional[int]
        The primary key of the instrument associated with this behavior session.
    trainer_pks : Optional[list[int]]
        The primary keys of the trainers associated with this behavior session.
    task : Optional[str]
        Name of the task associated with the session.
    task_stage : Optional[str]
        Name of the stage associated with the session.
    task_schema_version : Optional[str]
        Version of the task schema.
    is_curriculum_suggestion : Optional[bool]
        Whether the session is a curriculum suggestion.
    date : Optional[datetime]
        Date of the suggestion.
    notes : Optional[str]
        Notes about the session.
    """

    pk: Optional[int] = Field(
        default=None,
        serialization_alias="cnvn_pk",
        validation_alias="cnvn_pk",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="cnvn_createdOn",
        validation_alias="cnvn_createdOn",
    )
    mouse_pk: Optional[int] = Field(
        default=None,
        serialization_alias="cnvn_fk_content",
        validation_alias="cnvn_fk_content",
        description=(
            "The primary key of the mouse associated with this behavior session."
        ),
    )  # used as reference to mouse
    notes: Optional[str] = Field(
        default=None,
        serialization_alias="cnvn_cf_notes",
        validation_alias="cnvn_cf_notes",
    )
    task_stage: Optional[str] = Field(
        default=None,
        serialization_alias="cnvn_cf_taskStage",
        validation_alias="cnvn_cf_taskStage",
    )
    instrument_pk: Optional[int] = Field(
        default=None,
        serialization_alias="cnvn_fk_instrument",
        validation_alias="cnvn_fk_instrument",
    )
    trainer_pks: Optional[list[int]] = Field(
        default=[],
        serialization_alias="cnvn_cf_fk_trainer",
        validation_alias="cnvn_cf_fk_trainer",
    )
    task: Optional[str] = Field(
        default=None,
        serialization_alias="cnvn_cf_task",
        validation_alias="cnvn_cf_task",
    )
    is_curriculum_suggestion: Optional[bool] = Field(
        default=None,
        serialization_alias="cnvn_cf_stageIsOnCurriculum",
        validation_alias="cnvn_cf_stageIsOnCurriculum",
    )
    task_schema_version: Optional[str] = Field(
        default=None,
        serialization_alias="cnvn_cf_taskSchemaVersion",
        validation_alias="cnvn_cf_taskSchemaVersion",
    )
    software_version: Optional[str] = Field(
        default=None,
        serialization_alias="cnvn_cf_softwareVersion",
        validation_alias="cnvn_cf_softwareVersion",
    )
    date: Optional[datetime] = Field(
        default=None,
        serialization_alias="cnvn_cf_scheduledDate",
        validation_alias="cnvn_cf_scheduledDate",
    )
    cnvn_fk_contentEventType: int = 10  # pk of Behavior Session ContentEvent
    _slims_table = "ContentEvent"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "cnvt_name": "Behavior Session",
    }



"""Models for abstractions around Slims records. Friendlier names and
 documenation on how things relate to each other in our specific instance.
"""

from aind_slims_api.models.attachment import SlimsAttachment
from aind_slims_api.models.behavior_session import SlimsBehaviorSession
from aind_slims_api.models.instrument import SlimsInstrument
from aind_slims_api.models.metadata import SlimsMetadataReference
from aind_slims_api.models.mouse import SlimsMouseContent
from aind_slims_api.models.unit import SlimsUnit
from aind_slims_api.models.user import SlimsUser
from aind_slims_api.models.waterlog_result import SlimsWaterlogResult
from aind_slims_api.models.waterlog_water_restriction import SlimsWaterRestrictionEvent

__all__ = [
    "SlimsAttachment",
    "SlimsBehaviorSession",
    "SlimsInstrument",
    "SlimsMouseContent",
    "SlimsUnit",
    "SlimsUser",
    "SlimsWaterlogResult",
    "SlimsWaterRestrictionEvent",
    "SlimsMetadataReference",
]



"""
Module defining operations to build EcephysSession.
"""

import logging
from typing import List, Optional
from pydantic import BaseModel
from aind_slims_api import SlimsClient
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.mouse import SlimsMouseContent
from aind_slims_api.models.ecephys_session import (
    SlimsMouseSessionResult,
    SlimsStreamsResult,
    SlimsStimulusEpochsResult,
    SlimsExperimentRunStepContent,
    SlimsExperimentRunStep,
    SlimsDomeModuleRdrc,
    SlimsRewardDeliveryRdrc,
    SlimsRewardSpoutsRdrc,
    SlimsGroupOfSessionsRunStep,
    SlimsMouseSessionRunStep,
)

logger = logging.getLogger(__name__)


class EcephysSession(BaseModel):
    """
    Pydantic model encapsulating all session-related responses.
    """

    session_group: SlimsExperimentRunStep
    session_result: Optional[SlimsMouseSessionResult]
    streams: Optional[List[SlimsStreamsResult]] = []
    stream_modules: Optional[List[SlimsDomeModuleRdrc]] = []
    reward_delivery: Optional[SlimsRewardDeliveryRdrc] = None
    reward_spouts: Optional[SlimsRewardSpoutsRdrc] = None
    stimulus_epochs: Optional[List[SlimsStimulusEpochsResult]] = []


def _process_session_steps(
    client: SlimsClient,
    group_run_step: SlimsGroupOfSessionsRunStep,
    session_run_steps: List[SlimsMouseSessionRunStep],
) -> List[EcephysSession]:
    """
    Process session run steps and encapsulate related data into EcephysSession objects.
    Iterates through each run step in the provided session run steps,
    gathers the necessary data, and creates a list of EcephysSession objects.

    Parameters
    ----------
    client : SlimsClient
        An instance of SlimsClient used to retrieve additional session data.
    group_run_step : SlimsGroupOfSessionsRunStep
        The group run step containing session metadata and run information.
    session_run_steps : List[SlimsMouseSessionRunStep]
        A list of individual session run steps to be processed and encapsulated.

    Returns
    -------
    List[EcephysSession]
        A list of EcephysSession objects containing the processed session data.

    """
    ecephys_sessions = []

    for step in session_run_steps:
        # retrieve session, streams, and epochs from Results table
        session = client.fetch_model(
            SlimsMouseSessionResult, experiment_run_step_pk=step.pk
        )
        streams = client.fetch_models(SlimsStreamsResult, mouse_session_pk=session.pk)
        stimulus_epochs = client.fetch_models(
            SlimsStimulusEpochsResult, mouse_session_pk=session.pk
        )

        # retrieve modules and reward info from ReferenceDataRecord table
        stream_modules = [
            client.fetch_model(SlimsDomeModuleRdrc, pk=stream_module_pk)
            for stream in streams
            if stream.stream_modules_pk
            for stream_module_pk in stream.stream_modules_pk
        ]

        reward_delivery = (
            client.fetch_model(SlimsRewardDeliveryRdrc, pk=session.reward_delivery_pk)
            if session.reward_delivery_pk
            else None
        )

        reward_spouts = (
            client.fetch_model(
                SlimsRewardSpoutsRdrc, pk=reward_delivery.reward_spouts_pk
            )
            if reward_delivery and reward_delivery.reward_spouts_pk
            else None
        )

        # encapsulate all info for a single session
        ecephys_session = EcephysSession(
            session_group=group_run_step,
            session_result=session,
            streams=streams or None,
            stream_modules=stream_modules or None,
            reward_delivery=reward_delivery,
            reward_spouts=reward_spouts,
            stimulus_epochs=stimulus_epochs or None,
        )
        ecephys_sessions.append(ecephys_session)

    return ecephys_sessions


def fetch_ecephys_sessions(
    client: SlimsClient, subject_id: str
) -> List[EcephysSession]:
    """
    Fetch and process all electrophysiology (ecephys) run steps for a given subject.
    Retrieves all electrophysiology sessions associated with the provided subject ID
    and returns a list of EcephysSession objects.

    Parameters
    ----------
    client : SlimsClient
        An instance of SlimsClient used to connect to the SLIMS API.
    subject_id : str
        The ID of the subject (mouse) for which to fetch electrophysiology session data.

    Returns
    -------
    List[EcephysSession]
        A list of EcephysSession objects containing data for each run step.

    Example
    -------
    >>> from aind_slims_api import SlimsClient
    >>> client = SlimsClient()
    >>> sessions = fetch_ecephys_sessions(client=client, subject_id="000000")
    """
    ecephys_sessions_list = []
    mouse = client.fetch_model(SlimsMouseContent, barcode=subject_id)
    content_runs = client.fetch_models(SlimsExperimentRunStepContent, mouse_pk=mouse.pk)

    for content_run in content_runs:
        try:
            # retrieves content step to find experimentrun_pk
            content_run_step = client.fetch_model(
                SlimsExperimentRunStep, pk=content_run.runstep_pk
            )

            # retrieve group and mouse sessions in the experiment run
            group_run_step = client.fetch_models(
                SlimsGroupOfSessionsRunStep,
                experimentrun_pk=content_run_step.experimentrun_pk,
            )
            session_run_steps = client.fetch_models(
                SlimsMouseSessionRunStep,
                experimentrun_pk=content_run_step.experimentrun_pk,
            )
            if group_run_step and session_run_steps:
                ecephys_sessions = _process_session_steps(
                    client=client,
                    group_run_step=group_run_step[0],
                    session_run_steps=session_run_steps,
                )
                ecephys_sessions_list.extend(ecephys_sessions)

        except SlimsRecordNotFound as e:
            logging.info(str(e))
            continue

    return ecephys_sessions_list



"""Base model for SLIMS records abstraction.
"""

import logging
from datetime import datetime
from typing import ClassVar, Optional

from pydantic import BaseModel, ValidationInfo, field_serializer, field_validator
from slims.internal import Column as SlimsColumn  # type: ignore

from aind_slims_api.models.utils import _find_unit_spec
from aind_slims_api.types import SLIMS_TABLES

logger = logging.getLogger(__name__)


class SlimsBaseModel(
    BaseModel,
    from_attributes=True,
    validate_assignment=True,
    populate_by_name=True,
):
    """Pydantic model to represent a SLIMS record.
    Subclass with fields matching those in the SLIMS record.

    For Quantities, specify acceptable units like so:

        class MyModel(SlimsBaseModel):
            myfield: Annotated[float | None, UnitSpec("g","kg")]

        Quantities will be serialized using the first unit passed

    Datetime fields will be serialized to an integer ms timestamp
    """

    pk: Optional[int] = None
    created_on: Optional[datetime] = None
    json_entity: Optional[dict] = None
    _slims_table: ClassVar[SLIMS_TABLES]
    # base filters for model fetch
    _base_fetch_filters: ClassVar[dict[str, str]] = {}

    @field_validator("*", mode="before")
    def _validate(cls, value, info: ValidationInfo):
        """Validates a field, accounts for Quantities"""
        if isinstance(value, SlimsColumn):
            if value.datatype == "QUANTITY" and info.field_name is not None:
                unit_spec = _find_unit_spec(cls.model_fields[info.field_name])
                if unit_spec is None:
                    msg = (
                        f'Quantity field "{info.field_name}"'
                        "must be annotated with a UnitSpec"
                    )
                    raise TypeError(msg)
                if value.unit not in unit_spec.units:
                    msg = (
                        f'Unexpected unit "{value.unit}" for field '
                        f"{info.field_name}, Expected {unit_spec.units}"
                    )
                    raise ValueError(msg)
            return value.value
        else:
            return value

    @field_serializer("*")
    def _serialize(self, field, info):
        """Serialize a field, accounts for Quantities and datetime"""
        unit_spec = _find_unit_spec(self.model_fields[info.field_name])
        if unit_spec and field is not None:
            quantity = {
                "amount": field,
                "unit_display": unit_spec.preferred_unit,
            }
            return quantity
        elif isinstance(field, datetime):
            return int(field.timestamp() * 10**3)
        else:
            return field

    # TODO: Add links - need Record.json_entity['links']['self']
    # TODO: Add Table - need Record.json_entity['tableName']



"""Utility functions for filtering and sorting data in the SLIMS API.
"""

import logging
from copy import deepcopy
from typing import Any, Optional, Type, TypeVar, get_type_hints

from slims.criteria import Criterion, Expression, Junction, equals

from aind_slims_api.models.base import SlimsBaseModel

logger = logging.getLogger(__name__)

SlimsBaseModelTypeVar = TypeVar("SlimsBaseModelTypeVar", bound=SlimsBaseModel)


def resolve_model_alias(
    model: Type[SlimsBaseModelTypeVar],
    attr_name: str,
) -> str:
    """Given a SlimsBaseModel object, resolve an alias name to the actual slims
     column name.

    Notes
    -----
    - Raises ValueError if the alias cannot be resolved
    - Resolves the validation alias for a given field name
    - If prefixed with `-` will return the validation alias prefixed with
        `-`
    """
    has_prefix = attr_name.startswith("-")
    _attr_name = attr_name.lstrip("-")
    for field_name, field_info in model.model_fields.items():
        if (
            field_name == _attr_name
            and field_info.validation_alias
            and isinstance(field_info.validation_alias, str)
        ):
            alias = field_info.validation_alias
            if has_prefix:
                return f"-{alias}"
            return alias
    else:
        raise ValueError(f"Cannot resolve alias for {attr_name} on {model}")


CriteriaType = TypeVar("CriteriaType", Junction, Expression)


def resolve_criteria(
    model_type: Type[SlimsBaseModelTypeVar],
    criteria: Type[CriteriaType],
) -> Type[CriteriaType]:
    """Resolves criterion field name to serialization alias in a criterion."""

    if isinstance(criteria, Junction):
        criteria.members = [
            resolve_criteria(model_type, sub_criteria)
            for sub_criteria in criteria.members
        ]
        return criteria
    else:  # Expression
        if criteria.criterion["fieldName"] == "isNaFilter":
            criteria.criterion["value"] = resolve_model_alias(
                model_type,
                criteria.criterion["value"],
            )
        else:
            criteria.criterion["fieldName"] = resolve_model_alias(
                model_type,
                criteria.criterion["fieldName"],
            )
        return criteria


def _validate_field_name(
    model_type: Type[SlimsBaseModelTypeVar],
    field_name: str,
) -> None:
    """Check if field_name is a field on a model. Raises an Attribute error if
    it is not.
    """
    field_type_map = get_type_hints(model_type)
    if field_name not in field_type_map:
        raise ValueError(f"{field_name} is not a field on {model_type}.")


def _validate_field_value(
    model_type: Type[SlimsBaseModelTypeVar],
    field_name: str,
    field_value: Any,
) -> None:
    """Check if field_value is a compatible with the type associated with that
    field. Raises a ValueError if it is not.
    """
    field_type_map = get_type_hints(model_type)
    field_type = field_type_map[field_name]
    if not isinstance(field_value, field_type):
        raise ValueError(
            f"{field_value} is incompatible with {field_type}"
            f" for field {field_name}"
        )


def validate_criteria(
    model_type: Type[SlimsBaseModelTypeVar], criteria: Type[CriteriaType]
) -> None:
    """Check if field_name is a field on a model. Raises a ValueError if it
     is not.

    Raises
    ------
    AttributeError
        If the field_name is not a field on the model_type.
    ValueError
        If the field_value is not compatible with the field type.

    Notes
    -----
    - TODO: Consider using pydantic's validation error or some other more
     appropriate exception?
    """
    if isinstance(criteria, Junction):
        for sub_criteria in criteria.members:
            validate_criteria(model_type, sub_criteria)
    else:  # Expression
        if criteria.criterion["fieldName"] == "isNaFilter":
            _validate_field_name(
                model_type,
                criteria.criterion["value"],
            )
        elif criteria.criterion["operator"] in ["inSet", "notInSet"]:
            _validate_field_name(
                model_type,
                criteria.criterion["fieldName"],
            )
            for value in criteria.criterion["value"]:
                _validate_field_value(
                    model_type,
                    criteria.criterion["fieldName"],
                    value,
                )
        elif criteria.criterion["operator"] == "betweenInclusive":
            _validate_field_name(
                model_type,
                criteria.criterion["fieldName"],
            )
            _validate_field_value(
                model_type,
                criteria.criterion["fieldName"],
                criteria.criterion["start"],
            )
            _validate_field_value(
                model_type,
                criteria.criterion["fieldName"],
                criteria.criterion["end"],
            )
        else:
            _validate_field_name(
                model_type,
                criteria.criterion["fieldName"],
            )
            _validate_field_value(
                model_type,
                criteria.criterion["fieldName"],
                criteria.criterion["value"],
            )


def resolve_filter_args(
    model: Type[SlimsBaseModelTypeVar],
    *args: Criterion,
    sort: list[str] = [],
    start: Optional[int] = None,
    end: Optional[int] = None,
    **kwargs,
) -> tuple[list[Criterion], list[str], Optional[int], Optional[int]]:
    """Validates filter arguments and resolves field name aliases to SLIMS API
    column names.
    """
    criteria = deepcopy(list(args))
    criteria.extend(map(lambda item: equals(item[0], item[1]), kwargs.items()))
    resolved_criteria: list[Criterion] = []
    for criterion in criteria:
        validate_criteria(model, criterion)
        resolved_criteria.append(resolve_criteria(model, criterion))
    resolved_sort = [resolve_model_alias(model, sort_key) for sort_key in sort]
    if start is not None and end is None or end is not None and start is None:
        raise ValueError("Must provide both start and end or neither for fetch.")
    return resolved_criteria, resolved_sort, start, end



"""Models for the metadata stored in SLIMS."""

from datetime import datetime
from typing import Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


class SlimsMetadataReference(SlimsBaseModel):
    """Model for an instance of the metadata reference in SLIMS. Metadata
     content is added to metadata references in the form of attachments.

    Examples
    --------
    >>> from aind_slims_api import SlimsClient
    >>> from aind_slims_api import models
    >>> client = SlimsClient()

    ### Read
    >>> metadata_reference = client.fetch_model(
    ...  models.SlimsMetadataReference,
    ...  name="323_EPHYS1_OPTO_20240212"
    ... )
    >>> attachments = client.fetch_attachments(metadata_reference)
    >>> metadata = client.fetch_attachment_content(attachments[0])
    >>> metadata.json()["rig_id"]
    '323_EPHYS1_OPTO_2024-02-12'

    ### Read latest attachment
    >>> latest_attachment = client.fetch_attachment(
    ...  metadata_reference,
    ... )

    ### Write
    >>> import json
    >>> attachment_pk = client.add_attachment_content(
    ...  metadata_reference,
    ...  "test_metadata_attachment_name",
    ...  json.dumps({"rig_id": "323_EPHYS1_OPTO_2024-02-12"})
    ... )
    """

    name: str = Field(
        ...,
        serialization_alias="rdrc_name",
        validation_alias="rdrc_name",
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="rdrc_pk",
        validation_alias="rdrc_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"



"""Custom exceptions for the AIND Slims API."""


class SlimsAPIException(Exception):
    """Base exception for the AIND Slims API."""


class SlimsRecordNotFound(SlimsAPIException):
    """Exception raised when a record is not found in the SLIMS database."""



"""Contains a model for an ecephys result stored in SLIMS."""

from datetime import datetime
from typing import Annotated, List, Optional, ClassVar
from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.models.utils import UnitSpec


class SlimsExperimentRunStepContent(SlimsBaseModel):
    """Model for a SLIMS ExperimentRunStepContent"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="xrsc_pk", validation_alias="xrsc_pk"
    )
    mouse_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xrsc_fk_content",
        validation_alias="xrsc_fk_content",
    )
    runstep_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xrsc_fk_experimentRunStep",
        validation_alias="xrsc_fk_experimentRunStep",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="xrsc_createdOn",
        validation_alias="xrsc_createdOn",
    )
    _slims_table = "ExperimentRunStepContent"


class SlimsExperimentRunStep(SlimsBaseModel):
    """Model for a Slims ExperimentRunStep"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="xprs_pk", validation_alias="xprs_pk"
    )
    name: Optional[str] = Field(
        default=None, serialization_alias="xprs_name", validation_alias="xprs_name"
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="xprs_createdOn",
        validation_alias="xprs_createdOn",
    )
    experimentrun_pk: Optional[int] = Field(
        default=None,
        serialization_alias="xprs_fk_experimentRun",
        validation_alias="xprs_fk_experimentRun",
    )
    _slims_table = "ExperimentRunStep"


class SlimsGroupOfSessionsRunStep(SlimsExperimentRunStep):
    """Model for a Slims ExperimentRunStep"""

    session_type: Optional[str] = Field(
        default=None,
        serialization_alias="xprs_cf_sessionType",
        validation_alias="xprs_cf_sessionType",
    )
    mouse_platform_name: Optional[str] = Field(
        default=None,
        serialization_alias="xprs_cf_mousePlatformName",
        validation_alias="xprs_cf_mousePlatformName",
    )
    active_mouse_platform: Optional[bool] = Field(
        default=None,
        serialization_alias="xprs_cf_activeMousePlatform",
        validation_alias="xprs_cf_activeMousePlatform",
    )
    # TODO: add device calibrations once we have an example
    # device_calibrations_attachment: Optional[str] = Field(
    #     default=None,
    #     serialization_alias="xprs_cf_deviceCalibrations",
    #     validation_alias="xprs_cf_deviceCalibrations"
    # )
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Group of Sessions",
    }


class SlimsMouseSessionRunStep(SlimsExperimentRunStep):
    """Model for a Slims ExperimentRunStep"""

    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "xprs_name": "Mouse Session",
    }


class SlimsStreamsResult(SlimsBaseModel):
    """Model for a SLIMS Result Streams"""

    test_label: str = Field(
        "Streams", serialization_alias="test_label", validation_alias="test_label"
    )
    mouse_session: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession_display",
        validation_alias="rslt_cf_fk_mouseSession_display",
    )
    stream_modalities: Optional[List] = Field(
        default=None,
        serialization_alias="rslt_cf_streamModalities",
        validation_alias="rslt_cf_streamModalities",
    )
    daq_names: Optional[List] = Field(
        default=[],
        serialization_alias="rslt_cf_daqNames",
        validation_alias="rslt_cf_daqNames",
    )
    camera_names: Optional[List] = Field(
        default=None,
        serialization_alias="rslt_cf_cameraNames2",
        validation_alias="rslt_cf_cameraNames2",
    )
    # Look up in ReferenceDataRecord table
    stream_modules_pk: Optional[List] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_injectionMaterial2",
        validation_alias="rslt_cf_fk_injectionMaterial2",
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_pk",
        validation_alias="rslt_pk",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    mouse_session_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession",
        validation_alias="rslt_cf_fk_mouseSession",
    )
    experiment_run_step_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_fk_experimentRunStep",
        validation_alias="rslt_fk_experimentRunStep",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    _slims_table = "Result"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "test_name": "test_ephys_in_vivo_recording_stream",
    }


# TODO: Do we also need to check StimulusEpochs in ReferenceData?
class SlimsStimulusEpochsResult(SlimsBaseModel):
    """Model for a SLIMS Result Stimulus Epochs"""

    test_label: str = Field(
        "Stimulus Epochs",
        serialization_alias="test_label",
        validation_alias="test_label",
    )
    mouse_session: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession_display",
        validation_alias="rslt_cf_fk_mouseSession_display",
    )
    stimulus_device_names: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_stimulusDeviceNames",
        validation_alias="rslt_cf_stimulusDeviceNames",
    )
    stimulus_name: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_stimulusDeviceNames",
        validation_alias="rslt_cf_stimulusDeviceNames",
    )
    stimulus_modalities: Optional[List] = Field(
        default=None,
        serialization_alias="rslt_cf_stimulusModalities2",
        validation_alias="rslt_cf_stimulusModalities2",
    )
    reward_consumed_during_epoch: Annotated[float | None, UnitSpec("ml")] = Field(
        default=None,
        serialization_alias="rslt_cf_rewardConsumedDuringEpoch",
        validation_alias="rslt_cf_rewardConsumedDuringEpoch",
    )
    led_name: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_lightEmittingDiodeName",
        validation_alias="rslt_cf_lightEmittingDiodeName",
    )
    led_excitation_power_mw: Annotated[float | None, UnitSpec("mW")] = Field(
        default=None,
        serialization_alias="rslt_cf_lightEmittingDiodeExcitationPower",
        validation_alias="rslt_cf_lightEmittingDiodeExcitationPower",
    )
    laser_name: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_laserName",
        validation_alias="rslt_cf_laserName",
    )
    laser_wavelength: Annotated[float | None, UnitSpec("nm")] = Field(
        default=None,
        serialization_alias="rslt_laserWavelength",
        validation_alias="rslt_laserWavelength",
    )
    laser_excitation_power: Annotated[float | None, UnitSpec("mW")] = Field(
        default=None,
        serialization_alias="rslt_cf_laserExcitationPower",
        validation_alias="rslt_cf_laserExcitationPower",
    )
    speaker_name: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_speakerName",
        validation_alias="rslt_cf_speakerName",
    )
    speaker_volume: Annotated[float | None, UnitSpec("dB")] = Field(
        default=None,
        serialization_alias="rslt_cf_speakerVolume",
        validation_alias="rslt_cf_speakerVolume",
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_pk",
        validation_alias="rslt_pk",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    mouse_session_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession",
        validation_alias="rslt_cf_fk_mouseSession",
    )
    experiment_run_step_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_fk_experimentRunStep",
        validation_alias="rslt_fk_experimentRunStep",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    _slims_table = "Result"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "test_name": "test_stimulus_epochs",
    }


class SlimsMouseSessionResult(SlimsBaseModel):
    """Model for a SLIMS Result Mouse Session"""

    test_label: str = Field(
        "Mouse Session", serialization_alias="test_label", validation_alias="test_label"
    )
    mouse_session_id: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_uniqueIdentifier",
        validation_alias="rslt_uniqueIdentifier",
    )
    mouse_session: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession_display",
        validation_alias="rslt_cf_fk_mouseSession_display",
    )
    weight_prior_g: Annotated[float | None, UnitSpec("g")] = Field(
        default=None,
        serialization_alias="rslt_cf_animalWeightPrior",
        validation_alias="rslt_cf_animalWeightPrior",
    )
    weight_post_g: Annotated[float | None, UnitSpec("g")] = Field(
        default=None,
        serialization_alias="rslt_cf_animalWeightPost",
        validation_alias="rslt_cf_animalWeightPost",
    )
    reward_consumed_vol: Annotated[float | None, UnitSpec("ml")] = Field(
        default=None,
        serialization_alias="rslt_cf_rewardConsumedvolume",
        validation_alias="rslt_cf_rewardConsumedvolume",
    )
    reward_delivery_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_rewardDelivery",
        validation_alias="rslt_cf_fk_rewardDelivery",
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_pk",
        validation_alias="rslt_pk",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    mouse_session_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession",
        validation_alias="rslt_cf_fk_mouseSession",
    )
    experiment_run_step_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_fk_experimentRunStep",
        validation_alias="rslt_fk_experimentRunStep",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    _slims_table = "Result"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "test_name": "test_session_information",
    }


class SlimsEphysInsertionResult(SlimsBaseModel):
    """Model for a SLIMS Ephys Insertion Result"""

    test_label: str = Field(
        "EPHYS - Insertion",
        serialization_alias="test_label",
        validation_alias="test_label",
    )
    mouse_session: Optional[str] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession_display",
        validation_alias="rslt_cf_fk_mouseSession_display",
    )
    mouse_session_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rslt_cf_fk_mouseSession",
        validation_alias="rslt_cf_fk_mouseSession",
    )
    dye: Optional[str] = Field(
        default=None, serialization_alias="rslt_cf_dye", validation_alias="rslt_cf_dye"
    )
    manipulator_x: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rslt_cf_manipulatorCoordinates",
        validation_alias="rslt_cf_manipulatorCoordinates",
    )
    manipulator_y: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rslt_cf_manipulatorY",
        validation_alias="rslt_cf_manipulatorY",
    )
    insertion_fail: Optional[bool] = Field(
        default=None,
        serialization_alias="rslt_cf_failed",
        validation_alias="rslt_cf_failed",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    _slims_table = "Result"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "test_name": "test_ephys_insertion",
    }


class SlimsDomeModuleRdrc(SlimsBaseModel):
    """Model for Dome Module Reference Data"""

    pk: Optional[int] = Field(
        default=None, serialization_alias="rdrc_pk", validation_alias="rdrc_pk"
    )
    implant_hole: Optional[float] = Field(
        default=None, serialization_alias="rdrc_cf_bsl", validation_alias="rdrc_cf_bsl"
    )
    assembly_name: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_assemblyName",
        validation_alias="rdrc_cf_assemblyName",
    )
    probe_name: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_probeName",
        validation_alias="rdrc_cf_probeName",
    )
    primary_targeted_structure: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_fk_primaryTargetedStructure_display",
        validation_alias="rdrc_cf_fk_primaryTargetedStructure_display",
    )
    secondary_targeted_structures: Optional[List] = Field(
        default=None,
        serialization_alias="rdrc_cf_fk_secondaryTargetedStructures2_display",
        validation_alias="rdrc_cf_fk_secondaryTargetedStructures2_display",
    )
    arc_angle: Annotated[float | None, UnitSpec("degree", "°")] = Field(
        default=None,
        serialization_alias="rdrc_cf_arcAngle",
        validation_alias="rdrc_cf_arcAngle",
    )
    module_angle: Annotated[float | None, UnitSpec("degree", "°")] = Field(
        default=None,
        serialization_alias="rdrc_cf_moduleAngle",
        validation_alias="rdrc_cf_moduleAngle",
    )
    rotation_angle: Annotated[float | None, UnitSpec("degree", "°")] = Field(
        default=None,
        serialization_alias="rdrc_cf_rotationAngle",
        validation_alias="rdrc_cf_rotationAngle",
    )
    coordinate_transform: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_manipulatorCalibrations_display",
        validation_alias="rdrc_cf_manipulatorCalibrations_display",
    )
    ccf_coordinate_ap: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetedCcfCoordinatesAp",
        validation_alias="rdrc_cf_targetedCcfCoordinatesAp",
    )
    ccf_coordinate_ml: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetedCcfCoordinatesMl",
        validation_alias="rdrc_cf_targetedCcfCoordinatesMl",
    )
    ccf_coordinate_dv: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetedCcfCoordinatesDv",
        validation_alias="rdrc_cf_targetedCcfCoordinatesDv",
    )
    ccf_version: Optional[float] = Field(
        default=None,
        serialization_alias="rdrc_cf_ccfVersion",
        validation_alias="rdrc_cf_ccfVersion",
    )
    bregma_target_ap: Annotated[
        float | None, UnitSpec("dm", "pm", "cm", "mm", "&mu;m", "nm", "m", "dam", "Tm")
    ] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetAp",
        validation_alias="rdrc_cf_targetAp",
    )
    bregma_target_ml: Annotated[
        float | None, UnitSpec("dm", "pm", "cm", "mm", "&mu;m", "nm", "m", "dam", "Tm")
    ] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetMl",
        validation_alias="rdrc_cf_targetMl",
    )
    bregma_target_dv: Annotated[
        float | None, UnitSpec("dm", "pm", "cm", "mm", "&mu;m", "nm", "m", "dam", "Tm")
    ] = Field(
        default=None,
        serialization_alias="rdrc_cf_targetDv",
        validation_alias="rdrc_cf_targetDv",
    )
    surface_z: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_surfaceZ",
        validation_alias="rdrc_cf_surfaceZ",
    )
    manipulator_x: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_manipulatorX",
        validation_alias="rdrc_cf_manipulatorX",
    )
    manipulator_y: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_manipulatory",
        validation_alias="rdrc_cf_manipulatory",
    )
    manipulator_z: Annotated[float | None, UnitSpec("&mu;m")] = Field(
        default=None,
        serialization_alias="rdrc_cf_manipulatorZ",
        validation_alias="rdrc_cf_manipulatorZ",
    )
    dye: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_fk_dye_display",
        validation_alias="rdrc_cf_fk_dye_display",
    )
    fiber_connections_pk: Optional[int] = Field(
        default=None,
        serialization_alias="rdrc_cf_fk_fiberConnections",
        validation_alias="rdrc_cf_fk_fiberConnections",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "rdty_name": "Dome Module",
    }


class SlimsFiberConnectionsRdrc(SlimsBaseModel):
    """Model for Fiber Connections Reference Data"""

    pk: Optional[int] = Field(serialization_alias="rdrc_pk", validation_alias="rdrc_pk")
    patch_cord_name: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_patchCordName",
        validation_alias="rdrc_cf_patchCordName",
    )
    patch_cord_output_power: Annotated[float | None, UnitSpec("&mu;W")] = Field(
        default=None,
        serialization_alias="rdrc_cf_patchCordOutputPower",
        validation_alias="rdrc_cf_patchCordOutputPower",
    )
    fiber_names: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_fiberNames",
        validation_alias="rdrc_cf_fiberNames",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "rdty_name": "Fiber Connections",
    }


class SlimsRewardDeliveryRdrc(SlimsBaseModel):
    """Model for Reward Delivery Reference Data"""

    pk: Optional[int] = Field(serialization_alias="rdrc_pk", validation_alias="rdrc_pk")
    reward_spouts_pk: Optional[int] = Field(
        default=[],
        serialization_alias="rdrc_cf_fk_rewardSpouts",
        validation_alias="rdrc_cf_fk_rewardSpouts",
    )
    # OPTIONS: "Water", or "Other (if Other, specify below)"
    reward_solution: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_rewardSolution",
        validation_alias="rdrc_cf_rewardSolution",
    )
    other_reward_solution: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_specifyrewardSolution",
        validation_alias="rdrc_cf_specifyrewardSolution",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "rdty_name": "Reward Delivery",
    }


class SlimsRewardSpoutsRdrc(SlimsBaseModel):
    """Model for Reward Spouts Reference Data"""

    pk: Optional[int] = Field(serialization_alias="rdrc_pk", validation_alias="rdrc_pk")
    spout_side: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_spoutSide",
        validation_alias="rdrc_cf_spoutSide",
    )
    starting_position: Optional[str] = Field(
        default=None,
        serialization_alias="rdrc_cf_startingPosition",
        validation_alias="rdrc_cf_startingPosition",
    )
    variable_position: Optional[bool] = Field(
        default=None,
        serialization_alias="rdrc_cf_variablePosition",
        validation_alias="rdrc_cf_variablePosition",
    )
    created_on: Optional[datetime] = Field(
        default=None,
        serialization_alias="rdrc_createdOn",
        validation_alias="rdrc_createdOn",
    )
    _slims_table = "ReferenceDataRecord"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "rdty_name": "Reward Spouts",
    }



"""Common types for the SLIMS API.
"""

from typing import Literal

# List of slims tables manually accessed, there are many more
SLIMS_TABLES = Literal[
    "Attachment",
    "Project",
    "Content",
    "ContentEvent",
    "Unit",
    "Result",
    "ReferenceDataRecord",
    "Test",
    "User",
    "Groups",
    "Instrument",
    "Unit",
    "ExperimentRunStepContent",
    "ExperimentRunStep",
]



"""Contains a model for the instrument content, and a method for fetching it"""

from datetime import datetime
from typing import Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


class SlimsInstrument(SlimsBaseModel):
    """Model for a SLIMS instrument record.

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> client = SlimsClient()
    >>> instrument = client.fetch_model(SlimsInstrument, name="323_EPHYS1_OPTO")
    """

    # can't use alias for this due to https://github.com/pydantic/pydantic/issues/5893
    name: str = Field(
        ...,
        serialization_alias="nstr_name",
        validation_alias="nstr_name",
        description="The name of the instrument",
    )
    pk: Optional[int] = Field(
        default=None,
        serialization_alias="nstr_pk",
        validation_alias="nstr_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="nstr_createdOn",
        validation_alias="nstr_createdOn",
    )
    _slims_table = "Instrument"

    # todo add more useful fields



"""Contains a model for a waterlog result stored in SLIMS."""

from datetime import datetime
from typing import Annotated, ClassVar, Optional

from pydantic import Field

from aind_slims_api import __version__
from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.models.utils import UnitSpec


class SlimsWaterlogResult(SlimsBaseModel):
    """Model for a SLIMS Waterlog Result, the daily water/weight records

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> from aind_slims_api import models
    >>> client = SlimsClient()
    >>> mouse = client.fetch_model(models.SlimsMouseContent, barcode="00000000")
    >>> test_pk = client.fetch_pk("Test", test_name="test_waterlog")

    Write waterlog result.
    >>> waterlog_result = client.add_model(
    ...  models.SlimsWaterlogResult(
    ...   mouse_pk=mouse.pk,
    ...   date=datetime(2021,1,1),
    ...   weight_g=20.0,
    ...   water_earned_ml=5.0,
    ...   water_supplement_delivered_ml=5.0,
    ...   water_supplement_recommended_ml=5.0,
    ...   total_water_ml=10.0,
    ...   comments="comments",
    ...   workstation="aibs-computer-id",
    ...   test_pk=test_pk,
    ...  )
    ... )

    Read a waterlog result.
    >>> waterlog_result = client.fetch_model(
    ...  models.SlimsWaterlogResult,
    ...  mouse_pk=mouse.pk,
    ... )
    >>> waterlog_result.weight_g
    20.0
    """

    date: datetime = Field(
        datetime.now(),
        serialization_alias="rslt_cf_datePerformed",
        validation_alias="rslt_cf_datePerformed",
    )
    operator: Optional[str] = Field(
        None,
        serialization_alias="rslt_cf_waterlogOperator",
        validation_alias="rslt_cf_waterlogOperator",
    )
    weight_g: Annotated[float | None, UnitSpec("g")] = Field(
        None,
        serialization_alias="rslt_cf_weight",
        validation_alias="rslt_cf_weight",
    )
    water_earned_ml: Annotated[float | None, UnitSpec("ml")] = Field(
        ...,
        serialization_alias="rslt_cf_waterEarned",
        validation_alias="rslt_cf_waterEarned",
    )
    water_supplement_delivered_ml: Annotated[float | None, UnitSpec("ml")] = Field(
        ...,
        serialization_alias="rslt_cf_waterSupplementDelivered",
        validation_alias="rslt_cf_waterSupplementDelivered",
    )
    water_supplement_recommended_ml: Annotated[float | None, UnitSpec("ml")] = Field(
        ...,
        serialization_alias="rslt_cf_waterSupplementRecommended",
        validation_alias="rslt_cf_waterSupplementRecommended",
    )
    total_water_ml: Annotated[float | None, UnitSpec("ml")] = Field(
        ...,
        serialization_alias="rslt_cf_totalWater",
        validation_alias="rslt_cf_totalWater",
    )
    comments: Optional[str] = Field(
        None,
        serialization_alias="rslt_comments",
        validation_alias="rslt_comments",
    )
    workstation: Optional[str] = Field(
        None,
        serialization_alias="rslt_cf_fk_workStation",
        validation_alias="rslt_cf_fk_workStation",
    )
    sw_source: str = Field(
        "aind-slims-api",
        serialization_alias="rslt_cf_swSource",
        validation_alias="rslt_cf_swSource",
    )
    sw_version: str = Field(
        __version__,
        serialization_alias="rslt_cf_swVersion",
        validation_alias="rslt_cf_swVersion",
    )
    pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_pk",
        validation_alias="rslt_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="rslt_createdOn",
        validation_alias="rslt_createdOn",
    )
    mouse_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_content",
        validation_alias="rslt_fk_content",
    )
    test_pk: Optional[int] = Field(
        None,
        serialization_alias="rslt_fk_test",
        validation_alias="rslt_fk_test",
    )

    _slims_table = "Result"

    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "test_name": "test_waterlog",
    }



""" Tests the generic SlimsBaseModel"""

import unittest
from datetime import datetime
from typing import Annotated, Optional

from pydantic import Field
from slims.internal import Column, Record

from aind_slims_api.core import SlimsBaseModel
from aind_slims_api.models.utils import UnitSpec


class TestSlimsModel(unittest.TestCase):
    """Example Test Class"""

    class TestModel(SlimsBaseModel, validate_assignment=True):
        """Test case"""

        datefield: datetime = None
        stringfield: str = None
        quantfield: Annotated[float, UnitSpec("um", "nm")] = None

    def test_string_field(self):
        """Test basic usage for SLIMS column to Model field"""
        obj = self.TestModel()
        obj.stringfield = Column(
            {
                "datatype": "STRING",
                "name": "stringfield",
                "value": "value",
            }
        )

        self.assertEqual(obj.stringfield, "value")

    def test_quantity_field(self):
        """Test validation/serialization of a quantity type, with unit"""
        obj = self.TestModel()
        obj.quantfield = Column(
            {
                "datatype": "QUANTITY",
                "name": "quantfield",
                "value": 28.28,
                "unit": "um",
            }
        )

        self.assertEqual(obj.quantfield, 28.28)

        serialized = obj.model_dump()["quantfield"]
        expected = {"amount": 28.28, "unit_display": "um"}

        self.assertEqual(serialized, expected)

    def test_quantity_wrong_unit(self):
        """Ensure you get an error with an unexpected unit"""
        obj = self.TestModel()
        with self.assertRaises(ValueError):
            obj.quantfield = Column(
                {
                    "datatype": "QUANTITY",
                    "name": "quantfield",
                    "value": 28.28,
                    "unit": "erg",
                }
            )

    def test_alias(self):
        """Test aliasing of fields"""

        class TestModelAlias(SlimsBaseModel):
            """model with field aliases"""

            field: str = Field(
                ...,
                serialization_alias="alias",
                validation_alias="alias",
            )
            pk: Optional[int] = Field(
                None,
                serialization_alias="cntn_pk",
                validation_alias="cntn_pk",
            )

        record = Record(
            json_entity={
                "columns": [
                    {
                        "datatype": "STRING",
                        "name": "alias",
                        "value": "value",
                    },
                    {
                        "datatype": "INTEGER",
                        "name": "cntn_pk",
                        "value": 1,
                    },
                ]
            },
            slims_api=None,
        )
        obj = TestModelAlias.model_validate(record)

        self.assertEqual(obj.field, "value")
        obj.field = "value2"
        self.assertEqual(obj.field, "value2")
        serialized = obj.model_dump(include="field", by_alias=True)
        expected = {"alias": "value2"}
        self.assertEqual(serialized, expected)

    def test_unitspec(self):
        """Test unitspec with no arguments"""
        self.assertRaises(ValueError, UnitSpec)


if __name__ == "__main__":
    unittest.main()



"""Model for a record in the Attachment table in SLIMS."""

from datetime import datetime
from typing import Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


class SlimsAttachment(SlimsBaseModel):
    """Model for a record in the Attachment table in SLIMS.

    Examples
    --------
    >>> from aind_slims_api import SlimsClient
    >>> from aind_slims_api import models
    >>> client = SlimsClient()
    >>> rig_metadata_attachment = client.fetch_model(
    ...  models.SlimsAttachment,
    ...  name="rig323_EPHYS1_OPTO_2024-02-12.json"
    ... )
    >>> rig_metadata = client.fetch_attachment_content(
    ...  rig_metadata_attachment
    ... ).json()
    >>> rig_metadata["rig_id"]
    '323_EPHYS1_OPTO_2024-02-12'
    """

    pk: int = Field(
        ...,
        serialization_alias="attm_pk",
        validation_alias="attm_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="attm_createdOn",
        validation_alias="attm_createdOn",
    )
    name: str = Field(
        ...,
        serialization_alias="attm_name",
        validation_alias="attm_name",
    )
    _slims_table = "Attachment"



"""Tests methods in mouse module"""

import json
import os
import unittest
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

from slims.internal import Record

from aind_slims_api.core import SlimsClient
from aind_slims_api.models.behavior_session import SlimsBehaviorSession
from aind_slims_api.models.instrument import SlimsInstrument
from aind_slims_api.models.mouse import SlimsMouseContent
from aind_slims_api.models.user import SlimsUser

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"


class TestBehaviorSession(unittest.TestCase):
    """Tests top level methods in mouse module"""

    example_client: SlimsClient
    example_response: list[Record]
    example_mouse_response: list[Record]
    example_behavior_sessions: list[SlimsBehaviorSession]
    example_mouse: SlimsMouseContent
    example_write_sessions_response: Record
    example_instrument: SlimsInstrument
    example_trainer: SlimsUser

    @classmethod
    def setUpClass(cls):
        """Load json files of expected responses from slims"""
        cls.example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        cls.example_response = [
            Record(json_entity=r, slims_api=cls.example_client.db.slims_api)
            for r in json.loads(
                (
                    RESOURCES_DIR
                    / (
                        "example_fetch_behavior_session_content_events_response"
                        ".json_entity.json"
                    )
                ).read_text()
            )
        ]
        cls.example_mouse_response = [
            Record(json_entity=r, slims_api=cls.example_client.db.slims_api)
            for r in json.loads(
                (RESOURCES_DIR / "example_fetch_mouse_response.json").read_text()
            )
        ]
        assert (
            len(cls.example_response) > 1
        ), "Example response must be greater than 1 for tests to work..."

        cls.example_instrument = SlimsInstrument(
            nstr_name="323_EPHYS1_OPTO",
            nstr_pk=1,
        )
        cls.example_trainer = SlimsUser(
            user_userName="ClarkR",
            user_pk=1,
        )
        cls.example_mouse = SlimsMouseContent(
            cntn_barCode="00000000",
            cntn_pk=1,
            cntn_cf_waterRestricted=False,
            cntn_cf_scientificPointOfContact=None,
            cntn_cf_baselineWeight=None,
        )
        cls.example_behavior_sessions = [
            SlimsBehaviorSession(
                cnvn_cf_notes="Test notes",
                cnvn_cf_taskStage="Test stage",
                cnvn_cf_task="Test task",
                cnvn_cf_scheduledDate=datetime(2021, 1, 2),
            ),
        ]
        cls.example_write_sessions_response = [
            Record(json_entity=r, slims_api=cls.example_client.db.slims_api)
            for r in json.loads(
                (
                    RESOURCES_DIR
                    / (
                        "example_write_behavior_session_content_events_response."
                        "json_entity.json"
                    )
                ).read_text()
            )
        ][0]

    @patch("slims.slims.Slims.fetch")
    def test_fetch_behavior_session_content_events_success(self, mock_fetch: MagicMock):
        """Test fetch_behavior_session_content_events when successful"""
        mock_fetch.return_value = self.example_response
        validated = self.example_client.fetch_models(
            SlimsBehaviorSession,
            mouse_pk=self.example_mouse.pk,
            sort="date",
        )
        self.assertEqual(
            [item.json_entity for item in self.example_response],
            [item.json_entity for item in validated],
        )

    @patch("slims.slims.Slims.fetch")
    def test_fetch_behavior_session_content_events_success_sort_list(
        self, mock_fetch: MagicMock
    ):
        """Test fetch_behavior_session_content_events when successful"""
        mock_fetch.return_value = self.example_response
        validated = self.example_client.fetch_models(
            SlimsBehaviorSession,
            mouse_pk=self.example_mouse.pk,
            sort=["date"],
        )
        self.assertEqual(
            [item.json_entity for item in self.example_response],
            [item.json_entity for item in validated],
        )


if __name__ == "__main__":
    unittest.main()



""" Tests methods in core module"""

import json
import os
import unittest
from copy import deepcopy
from pathlib import Path
from unittest.mock import MagicMock, patch

from requests import Response
from slims.criteria import conjunction, equals
from slims.internal import Record, _SlimsApiException

from aind_slims_api.core import SlimsAttachment, SlimsClient
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.behavior_session import SlimsBehaviorSession
from aind_slims_api.models.unit import SlimsUnit
from aind_slims_api.models.user import SlimsUser

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"


class TestSlimsClient(unittest.TestCase):
    """Tests methods in SlimsClient class"""

    example_client: SlimsClient
    example_fetch_unit_response: list[Record]
    example_fetch_mouse_response: list[Record]
    example_fetch_user_response: list[Record]
    example_fetch_attachment_response: list[Record]
    example_add_attachments_response_text: str

    @classmethod
    def setUpClass(cls):
        """Sets up class by downloading responses"""
        example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        cls.example_client = example_client

        def get_response(attribute_name: str):
            """Utility method to download json file with {attribute_name}.json
            from resource dir"""
            with open(
                os.path.join(str(RESOURCES_DIR), f"{attribute_name}.json"), "r"
            ) as f:
                response = [
                    Record(json_entity=r, slims_api=example_client.db.slims_api)
                    for r in json.load(f)
                ]
            return response

        cls.example_fetch_unit_response = get_response("example_fetch_unit_response")
        cls.example_fetch_mouse_response = get_response("example_fetch_mouse_response")
        cls.example_fetch_user_response = get_response("example_fetch_user_response")
        cls.example_fetch_attachment_response = get_response(
            "example_fetch_attachments_response.json_entity"
        )
        cls.example_add_attachments_response_text = (
            RESOURCES_DIR / "example_add_attachment_content_response_text.txt"
        ).read_text()

    def test_rest_link(self):
        """Tests rest_link method with both queries and no queries."""

        rest_link_no_queries = self.example_client.rest_link(table="Content")
        rest_link_with_queries = self.example_client.rest_link(
            table="Content", **{"limit": 1, "start": 0}
        )
        self.assertEqual("http://fake_url/rest/Content", rest_link_no_queries)
        self.assertEqual(
            "http://fake_url/rest/Content?limit=1?start=0", rest_link_with_queries
        )

    @patch("slims.slims.Slims.fetch")
    def test_fetch(self, mock_slims_fetch: MagicMock):
        """Tests fetch method success"""
        mock_slims_fetch.return_value = self.example_fetch_unit_response
        response = self.example_client.fetch(table="Unit", start=0, end=2)
        self.assertEqual(self.example_fetch_unit_response, response)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_with_criteria(self, mock_slims_fetch: MagicMock):
        """Tests fetch method constructs criteria correctly"""
        mock_slims_fetch.return_value = self.example_fetch_mouse_response
        response = self.example_client.fetch(
            "Content",
            equals("cntn_barCode", "123456"),
            cntp_name="Mouse",
        )
        expected_criteria = (
            conjunction()
            .add(equals("cntn_barCode", "123456"))
            .add(equals("cntp_name", "Mouse"))
        )
        actual_criteria = mock_slims_fetch.mock_calls[0].args[1]
        self.assertEqual(expected_criteria.to_dict(), actual_criteria.to_dict())
        self.assertEqual(self.example_fetch_mouse_response, response)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_error(self, mock_slims_fetch: MagicMock):
        """Tests fetch method when a _SlimsApiException is raised"""
        mock_slims_fetch.side_effect = _SlimsApiException("Something went wrong")
        with self.assertRaises(_SlimsApiException) as e:
            self.example_client.fetch(
                "Content",
                cntp_name="Mouse",
            )
        self.assertEqual("Something went wrong", e.exception.args[0])

    @patch("slims.slims.Slims.fetch")
    def test_fetch_user(self, mock_slims_fetch: MagicMock):
        """Tests fetch_user method"""
        mock_slims_fetch.return_value = self.example_fetch_user_response
        response = self.example_client.fetch_user(user_name="PersonA")
        self.assertEqual(self.example_fetch_user_response, response)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_pk(self, mock_slims_fetch: MagicMock):
        """Tests fetch_pk method when several records are returned"""
        # Use this example_client since result is cached
        example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        mock_slims_fetch.return_value = self.example_fetch_unit_response
        pk = example_client.fetch_pk(table="Unit")
        self.assertEqual(31, pk)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_pk_none(self, mock_slims_fetch: MagicMock):
        """Tests fetch_pk method when no records are returned"""
        # Use this example_client since result is cached
        example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        mock_slims_fetch.return_value = []
        pk = example_client.fetch_pk(table="Content")
        self.assertIsNone(pk)

    @patch("logging.Logger.info")
    @patch("slims.slims.Slims.add")
    def test_add(self, mock_slims_add: MagicMock, mock_log: MagicMock):
        """Tests add method"""
        mock_slims_add.return_value = self.example_fetch_unit_response[0]
        input_data = deepcopy(self.example_fetch_unit_response[0].json_entity)
        record = self.example_client.add(table="Unit", data=input_data)
        self.assertEqual(self.example_fetch_unit_response[0], record)
        mock_log.assert_called_once_with("SLIMS Add: Unit/31")

    @patch("slims.slims.Slims.fetch_by_pk")
    @patch("logging.Logger.info")
    @patch("slims.internal.Record.update")
    def test_update(
        self,
        mock_update: MagicMock,
        mock_log: MagicMock,
        mock_fetch_by_pk: MagicMock,
    ):
        """Tests update method success"""
        input_data = deepcopy(self.example_fetch_unit_response[0].json_entity)
        mock_record = Record(
            json_entity=input_data, slims_api=self.example_client.db.slims_api
        )
        mock_fetch_by_pk.return_value = mock_record
        new_data = deepcopy(input_data)
        new_data["columns"][0]["value"] = "PM^3"
        mocked_updated_record = Record(
            json_entity=new_data, slims_api=self.example_client.db.slims_api
        )
        mock_update.return_value = mocked_updated_record
        new_record = self.example_client.update(table="Unit", pk=31, data=new_data)
        self.assertEqual(mocked_updated_record, new_record)
        mock_log.assert_called_once_with("SLIMS Update: Unit/31")

    @patch("slims.slims.Slims.fetch_by_pk")
    @patch("logging.Logger.info")
    @patch("slims.internal.Record.update")
    def test_update_failure(
        self,
        mock_update: MagicMock,
        mock_log: MagicMock,
        mock_fetch_by_pk: MagicMock,
    ):
        """Tests update method when a failure occurs"""
        mock_fetch_by_pk.return_value = None
        with self.assertRaises(ValueError) as e:
            self.example_client.update(table="Unit", pk=30000, data={})
        self.assertEqual(
            'No data in SLIMS "Unit" table for pk "30000"',
            e.exception.args[0],
        )
        mock_update.assert_not_called()
        mock_log.assert_not_called()

    @patch("slims.slims.Slims.fetch_by_pk")
    @patch("logging.Logger.info")
    @patch("slims.internal.Record.update")
    def test_update_model_no_pk(
        self,
        mock_update: MagicMock,
        mock_log: MagicMock,
        mock_fetch_by_pk: MagicMock,
    ):
        """Tests update method when a failure occurs"""
        mock_fetch_by_pk.return_value = None
        with self.assertRaises(ValueError):
            self.example_client.update_model(SlimsUnit.model_construct(pk=None))
        mock_update.assert_not_called()
        mock_log.assert_not_called()

    @patch("logging.Logger.info")
    @patch("slims.slims.Slims.add")
    def test_add_model(self, mock_slims_add: MagicMock, mock_log: MagicMock):
        """Tests add_model method with mock mouse data"""
        record = self.example_fetch_unit_response[0]
        mock_slims_add.return_value = record
        input_model = SlimsUnit.model_validate(record)
        added = self.example_client.add_model(input_model)
        self.assertEqual(input_model, added)
        mock_log.assert_called_once_with("SLIMS Add: Unit/31")

    @patch("slims.slims.Slims.fetch_by_pk")
    @patch("logging.Logger.info")
    @patch("slims.internal.Record.update")
    def test_update_model(
        self,
        mock_update: MagicMock,
        mock_log: MagicMock,
        mock_fetch_by_pk: MagicMock,
    ):
        """Tests update method success"""
        input_data = deepcopy(self.example_fetch_unit_response[0].json_entity)
        mock_record = Record(
            json_entity=input_data, slims_api=self.example_client.db.slims_api
        )
        mock_fetch_by_pk.return_value = mock_record
        updated_model = SlimsUnit.model_validate(mock_record)
        new_data = deepcopy(input_data)
        new_data["columns"][0]["value"] = "PM^3"
        mocked_updated_record = Record(
            json_entity=new_data, slims_api=self.example_client.db.slims_api
        )
        mock_update.return_value = mocked_updated_record
        updated_model = SlimsUnit.model_validate(mocked_updated_record)
        returned_model = self.example_client.update_model(updated_model)
        self.assertEqual(updated_model, returned_model)
        mock_log.assert_called_once_with("SLIMS Update: Unit/31")

    @patch("slims.slims.Slims.fetch")
    def test_fetch_model_no_records(self, mock_slims_fetch: MagicMock):
        """Tests fetch_user method"""
        mock_slims_fetch.return_value = []
        with self.assertRaises(SlimsRecordNotFound):
            self.example_client.fetch_model(SlimsUnit)

    @patch("slims.internal._SlimsApi.get_entities")
    def test_fetch_attachments(self, mock_get_entities: MagicMock):
        """Tests fetch_attachments method success."""
        # slims_api is dynamically added to slims client
        mock_get_entities.return_value = self.example_fetch_attachment_response
        unit = SlimsUnit.model_validate(
            Record(
                json_entity=self.example_fetch_unit_response[0].json_entity,
                slims_api=self.example_client.db.slims_api,
            )
        )
        attachments = self.example_client.fetch_attachments(
            unit,
        )
        self.assertEqual(1, len(attachments))

    @patch("slims.slims._SlimsApi.get")
    def test_fetch_attachment_content(self, mock_get: MagicMock):
        """Tests fetch_attachment_content method success."""
        mock_get.return_value = Response()
        self.example_client.fetch_attachment_content(
            SlimsAttachment(
                attm_name="test",
                attm_pk=1,
            )
        )
        mock_get.assert_called_once()

    @patch("slims.internal._SlimsApi.post")
    def test_add_attachment_content(self, mock_post: MagicMock):
        """Tests add_attachment_content method success."""
        mock_response = MagicMock()
        mock_response.text = self.example_add_attachments_response_text
        mock_post.return_value = mock_response
        unit_pk = 1
        result = self.example_client.add_attachment_content(
            SlimsUnit(
                unit_name="test",
                unit_pk=unit_pk,
            ),
            "test",
            "some test content",
        )
        self.assertEqual(result, int(self.example_add_attachments_response_text))
        self.assertEqual(mock_post.call_count, 1)
        self.assertEqual(
            mock_post.mock_calls[0].kwargs["body"]["atln_recordPk"], unit_pk
        )

    @patch("slims.internal._SlimsApi.post")
    def test_add_attachment_content_no_pk(self, mock_post: MagicMock):
        """Tests add_attachment_content method failure due to lack of pk."""
        mock_response = MagicMock()
        mock_response.text.return_value = self.example_add_attachments_response_text
        mock_post.return_value = mock_response
        with self.assertRaises(ValueError):
            self.example_client.add_attachment_content(
                SlimsBehaviorSession(),
                "test",
                "some test content",
            )
        mock_post.assert_not_called()

    @patch("logging.Logger.error")
    def test__validate_model_invalid_model(self, mock_log: MagicMock):
        """Tests _validate_model method with one invalid model and one valid
        one.
        """
        valid_data = deepcopy(self.example_fetch_unit_response[0].json_entity)
        invalid_data = deepcopy(self.example_fetch_unit_response[0].json_entity)
        invalid_data["columns"][0]["value"] = 1
        validated = self.example_client._validate_models(
            SlimsUnit,
            [
                Record(
                    json_entity=valid_data,
                    slims_api=self.example_client.db.slims_api,
                ),
                Record(
                    json_entity=invalid_data,
                    slims_api=self.example_client.db.slims_api,
                ),
            ],
        )
        self.assertEqual(1, len(validated))
        self.assertEqual(1, mock_log.call_count)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_model_criterion(self, mock_slims_fetch: MagicMock):
        """Tests fetch_model method with a criterion."""
        mock_slims_fetch.return_value = self.example_fetch_user_response
        self.example_client.fetch_model(SlimsUser, equals("username", "LKim"))
        mock_slims_fetch.assert_called_once()

    @patch("slims.slims.Slims.fetch")
    def test_fetch_model_criterion_junction(self, mock_slims_fetch: MagicMock):
        """Tests fetch_model method with a junction criterion."""
        mock_slims_fetch.return_value = self.example_fetch_user_response
        self.example_client.fetch_model(
            SlimsUser,
            conjunction().add(equals("username", "LKim")),
        )
        mock_slims_fetch.assert_called_once()

    @patch("slims.slims.Slims.fetch")
    def test_fetch_model_criterion_invalid_criterion_type(
        self, mock_slims_fetch: MagicMock
    ):
        """Tests fetch_model method with an invalid criterion type."""
        mock_slims_fetch.return_value = []
        with self.assertRaises(ValueError):
            self.example_client.fetch_model(SlimsUser, equals("username", 1))
        mock_slims_fetch.assert_not_called()

    @patch("slims.internal._SlimsApi.get_entities")
    def test_fetch_attachment(self, mock_get_entities: MagicMock):
        """Tests fetch_attachment method success."""
        mock_get_entities.return_value = self.example_fetch_attachment_response
        unit = SlimsUnit.model_validate(
            Record(
                json_entity=self.example_fetch_unit_response[0].json_entity,
                slims_api=self.example_client.db.slims_api,
            )
        )
        self.example_client.fetch_attachment(unit, equals("name", "test"))

    @patch("slims.internal._SlimsApi.get_entities")
    def test_fetch_attachment_no_attachments(self, mock_get_entities: MagicMock):
        """Tests fetch_attachment method failure due to no attachments."""
        mock_get_entities.return_value = []
        unit = SlimsUnit.model_validate(
            Record(
                json_entity=self.example_fetch_unit_response[0].json_entity,
                slims_api=self.example_client.db.slims_api,
            )
        )
        with self.assertRaises(SlimsRecordNotFound):
            self.example_client.fetch_attachment(unit)

    @patch("slims.slims.Slims.fetch")
    def test_fetch_str_sort(self, mock_slims_fetch: MagicMock):
        """Tests fetch method when sort is a string."""
        mock_slims_fetch.return_value = []
        self.example_client.fetch(SlimsUser._slims_table, sort="username")
        mock_slims_fetch.assert_called_once()

    @patch("slims.slims.Slims.fetch")
    def test_fetch_models_invalid_start_end(self, mock_slims_fetch: MagicMock):
        """Tests fetch_model method failure due to only supplying start or end."""
        mock_slims_fetch.return_value = []
        with self.assertRaises(ValueError):
            self.example_client.fetch_models(SlimsUser, start=1)
        with self.assertRaises(ValueError):
            self.example_client.fetch_models(SlimsUser, end=1)
        mock_slims_fetch.assert_not_called()


if __name__ == "__main__":
    unittest.main()



"""Testing ecephys session operation"""

import os
import unittest
import json
from unittest.mock import patch
from pathlib import Path
from slims.internal import Record
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.models.mouse import SlimsMouseContent
from aind_slims_api.models.ecephys_session import (
    SlimsMouseSessionResult,
    SlimsStreamsResult,
    SlimsDomeModuleRdrc,
    SlimsRewardDeliveryRdrc,
    SlimsRewardSpoutsRdrc,
    SlimsGroupOfSessionsRunStep,
    SlimsMouseSessionRunStep,
    SlimsExperimentRunStepContent,
    SlimsExperimentRunStep,
)
from aind_slims_api.operations import EcephysSession, fetch_ecephys_sessions

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / ".." / "resources"


class TestSlimsEcephysSessionOperator(unittest.TestCase):
    """Test class for SlimsEcephysSessionOperator"""

    @patch("aind_slims_api.operations.ecephys_session.SlimsClient")
    def setUp(cls, mock_client):
        """setup test class"""
        cls.mock_client = mock_client()
        with open(
            RESOURCES_DIR / "example_fetch_ecephys_session_result.json", "r"
        ) as f:
            response = [
                Record(json_entity=r, slims_api=cls.mock_client.db.slims_api)
                for r in json.load(f)
            ]
        cls.example_fetch_ecephys_session_result = response

    def test_fetch_ecephys_sessions_success(self):
        """Tests session info is fetched successfully"""
        self.mock_client.fetch_models.side_effect = [
            [SlimsExperimentRunStepContent(pk=1, runstep_pk=3, mouse_pk=12345)],
            [
                SlimsGroupOfSessionsRunStep(
                    pk=6,
                    session_type="OptoTagging",
                    mouse_platform_name="Platform1",
                    experimentrun_pk=101,
                )
            ],
            [SlimsMouseSessionRunStep(pk=7, experimentrun_pk=101)],
            [
                SlimsStreamsResult(
                    pk=8,
                    mouse_session_pk=7,
                    stream_modules_pk=[9, 10],
                    daq_names=["DAQ1", "DAQ2"],
                )
            ],
            [],
        ]

        self.mock_client.fetch_model.side_effect = [
            SlimsMouseContent.model_construct(pk=12345),
            SlimsExperimentRunStep(pk=3, experimentrun_pk=101),
            SlimsMouseSessionResult(pk=12, reward_delivery_pk=14),
            SlimsDomeModuleRdrc(pk=9, probe_name="Probe1", arc_angle=20),
            SlimsDomeModuleRdrc(pk=10, probe_name="Probe1", arc_angle=20),
            SlimsRewardDeliveryRdrc(
                pk=3, reward_spouts_pk=5, reward_solution="Solution1"
            ),
            SlimsRewardSpoutsRdrc(pk=5, spout_side="Left"),
        ]

        # Run the fetch_sessions method
        ecephys_sessions = fetch_ecephys_sessions(
            client=self.mock_client, subject_id="12345"
        )

        # Assertions
        self.assertEqual(len(ecephys_sessions), 1)
        ecephys_session = ecephys_sessions[0]
        self.assertIsInstance(ecephys_session, EcephysSession)
        self.assertEqual(ecephys_session.session_group.session_type, "OptoTagging")
        self.assertEqual(len(ecephys_session.streams), 1)
        self.assertEqual(ecephys_session.streams[0].daq_names, ["DAQ1", "DAQ2"])
        self.assertEqual(len(ecephys_session.stream_modules), 2)
        self.assertIsNone(ecephys_session.stimulus_epochs)

    def test_fetch_ecephys_sessions_handle_exception(self):
        """Tests that exception is handled as expected"""
        self.mock_client.fetch_models.side_effect = [
            [SlimsExperimentRunStepContent(pk=1, runstep_pk=3, mouse_pk=67890)]
        ]
        self.mock_client.fetch_model.side_effect = [
            SlimsMouseContent.model_construct(pk=67890),
            SlimsRecordNotFound("No record found for SlimsExperimentRunStep with pk=3"),
        ]

        with patch("logging.info") as mock_log_info:
            fetch_ecephys_sessions(client=self.mock_client, subject_id="67890")
            mock_log_info.assert_called_with(
                "No record found for SlimsExperimentRunStep with pk=3"
            )


if __name__ == "__main__":
    unittest.main()



"""Contains a model for a user."""

from datetime import datetime
from typing import Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


# TODO: Tighten this up once users are more commonly used
class SlimsUser(SlimsBaseModel):
    """Model for user information in SLIMS

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> client = SlimsClient()
    >>> user = client.fetch_model(SlimsUser, username="LKim")
    """

    username: str = Field(
        ...,
        serialization_alias="user_userName",
        validation_alias="user_userName",
    )
    first_name: Optional[str] = Field(
        "",
        serialization_alias="user_firstName",
        validation_alias="user_firstName",
    )
    last_name: Optional[str] = Field(
        "",
        serialization_alias="user_lastName",
        validation_alias="user_lastName",
    )
    full_name: Optional[str] = Field(
        "",
        serialization_alias="user_fullName",
        validation_alias="user_fullName",
    )
    email: Optional[str] = Field(
        "",
        serialization_alias="user_email",
        validation_alias="user_email",
    )
    pk: int = Field(
        ...,
        serialization_alias="user_pk",
        validation_alias="user_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="user_createdOn",
        validation_alias="user_createdOn",
    )

    _slims_table = "User"



"""Contents:

Utilities for creating pydantic models for SLIMS data:
    SlimsBaseModel - to be subclassed for SLIMS pydantic models
    UnitSpec - To be included in a type annotation of a Quantity field

SlimsClient - Basic wrapper around slims-python-api client with convenience
    methods and integration with SlimsBaseModel subtypes
"""

import base64
import logging
from copy import deepcopy
from functools import lru_cache
from typing import Any, Optional, Type, TypeVar

from pydantic import ValidationError
from requests import Response
from slims.criteria import Criterion, conjunction, equals
from slims.internal import Record as SlimsRecord
from slims.slims import Slims, _SlimsApiException

from aind_slims_api import config
from aind_slims_api.exceptions import SlimsRecordNotFound
from aind_slims_api.filters import resolve_filter_args
from aind_slims_api.models import SlimsAttachment
from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.types import SLIMS_TABLES

logger = logging.getLogger(__name__)


SlimsBaseModelTypeVar = TypeVar("SlimsBaseModelTypeVar", bound=SlimsBaseModel)


class SlimsClient:
    """Wrapper around slims-python-api client with convenience methods"""

    db: Slims

    def __init__(self, url=None, username=None, password=None):
        """Create object and try to connect to database"""
        self.url = url or config.slims_url

        self.connect(
            self.url,
            username or config.slims_username,
            password or config.slims_password.get_secret_value(),
        )

    def connect(self, url: str, username: str, password: str):
        """Connect to the database"""
        self.db = Slims(
            "slims",
            url,
            username,
            password,
        )

    def fetch(
        self,
        table: SLIMS_TABLES,
        *args,
        sort: Optional[str | list[str]] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsRecord]:
        """Fetch from the SLIMS database

        Args:
            table (str): SLIMS table to query
            sort (str | list[str], optional): Fields to sort by; e.g. date
            start (int, optional):  The first row to return
            end (int, optional): The last row to return
            *args (Slims.criteria.Criterion): Optional criteria to apply
            **kwargs (dict[str,str]): "field=value" filters

        Returns:
            records (list[SlimsRecord] | None): Matching records, if any
        """
        criteria = conjunction()
        for arg in args:
            if isinstance(arg, Criterion):
                criteria.add(arg)

        for k, v in kwargs.items():
            criteria.add(equals(k, v))

        if isinstance(sort, str):
            sort = [sort]

        try:
            records = self.db.fetch(
                table,
                criteria,
                sort=sort,
                start=start,
                end=end,
            )
        except _SlimsApiException as e:
            # TODO: Add better error handling
            #  Let's just raise error for the time being
            raise e

        return records

    @staticmethod
    def _validate_models(
        model_type: Type[SlimsBaseModelTypeVar], records: list[SlimsRecord]
    ) -> list[SlimsBaseModelTypeVar]:
        """Validate a list of SlimsBaseModel objects. Logs errors for records
        that fail pydantic validation."""
        validated = []
        for record in records:
            try:
                validated.append(model_type.model_validate(record))
            except ValidationError as e:
                logger.error(f"SLIMS data validation failed, {repr(e)}")
        return validated

    def fetch_models(
        self,
        model: Type[SlimsBaseModelTypeVar],
        *args: Criterion,
        sort: str | list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsBaseModelTypeVar]:
        """Fetch records from SLIMS and return them as SlimsBaseModel objects.

        Returns
        -------
        tuple:
            list:
                Validated SlimsBaseModel objects

        Notes
        -----
        - kwargs are mapped to field alias names and used as equality filters
         for the fetch.
        """
        resolved_kwargs = deepcopy(model._base_fetch_filters)
        logger.debug("Resolved kwargs: %s", resolved_kwargs)

        if isinstance(sort, str):
            sort = [sort]

        criteria, resolved_sort, start, end = resolve_filter_args(
            model,
            *args,
            sort=sort,
            start=start,
            end=end,
            **kwargs,
        )
        response = self.fetch(
            model._slims_table,
            *criteria,
            sort=resolved_sort,
            start=start,
            end=end,
            **resolved_kwargs,
        )
        return self._validate_models(model, response)

    def fetch_model(
        self,
        model: Type[SlimsBaseModelTypeVar],
        *args: Criterion,
        **kwargs,
    ) -> SlimsBaseModelTypeVar | None:
        """Fetch a single record from SLIMS and return it as a validated
         SlimsBaseModel object.

        Notes
        -----
        - kwargs are mapped to field alias values
        - sorts records on created_on in descending order and returns the first
        """
        records = self.fetch_models(
            model,
            *args,
            sort="-created_on",
            start=0,  # slims rows appear to be 0-indexed
            end=1,
            **kwargs,
        )
        if len(records) > 0:
            logger.debug(f"Found {len(records)} records for {model}.")
        if len(records) < 1:
            raise SlimsRecordNotFound(f"No record found for {model} with {args}.")
        return records[0]

    @staticmethod
    def _create_get_entities_body(
        *args: Criterion,
        sort: list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
    ) -> dict[str, Any]:
        """Creates get entities body for SLIMS API request."""
        body: dict[str, Any] = {
            "sortBy": sort,
            "startRow": start,
            "endRow": end,
        }
        if args:
            criteria = conjunction()
            for arg in args:
                criteria.add(arg)
            body["criteria"] = criteria.to_dict()

        return body

    def fetch_attachments(
        self,
        record: SlimsBaseModel,
        *args: Criterion,
        sort: str | list[str] = [],
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ) -> list[SlimsAttachment]:
        """Fetch attachments for a given record.

        Notes
        -----
        - kwargs are mapped to field alias values
        """
        if isinstance(sort, str):
            sort = [sort]

        criteria, sort, start, end = resolve_filter_args(
            SlimsAttachment,
            *args,
            sort=sort,
            start=start,
            end=end,
            **kwargs,
        )
        return self._validate_models(
            SlimsAttachment,
            self.db.slims_api.get_entities(
                f"attachment/{record._slims_table}/{record.pk}",
                body=self._create_get_entities_body(
                    *criteria,
                    sort=sort,
                    start=start,
                    end=end,
                ),
            ),
        )

    def fetch_attachment(
        self,
        record: SlimsBaseModel,
        *args: Criterion,
        **kwargs,
    ) -> SlimsAttachment:
        """Fetch attachments for a given record.

        Notes
        -----
        - kwargs are mapped to field alias values
        - sorts records on created_on in descending order and returns the first
        """
        records = self.fetch_attachments(
            record,
            *args,
            sort="-created_on",
            start=0,  # slims rows appear to be 0-indexed
            end=1,
            **kwargs,
        )
        if len(records) > 0:
            logger.debug(f"Found {len(records)} records for {record}.")
        if len(records) < 1:
            raise SlimsRecordNotFound("No record found.")
        return records[0]

    def fetch_attachment_content(
        self,
        attachment: int | SlimsAttachment,
    ) -> Response:
        """Fetch attachment content for a given attachment.

        Parameters
        -----------
        attachment: int | SlimsAttachment
            The primary key of the attachment or an attachment object
        """
        if isinstance(attachment, SlimsAttachment):
            attachment = attachment.pk

        return self.db.slims_api.get(f"repo/{attachment}")

    @lru_cache(maxsize=None)
    def fetch_pk(self, table: SLIMS_TABLES, *args, **kwargs) -> int | None:
        """SlimsClient.fetch but returns the pk of the first returned record"""
        records = self.fetch(table, *args, **kwargs)
        if len(records) > 0:
            return records[0].pk()
        else:
            return None

    def fetch_user(self, user_name: str):
        """Fetches a user by username"""
        return self.fetch("User", user_userName=user_name)

    def add(self, table: SLIMS_TABLES, data: dict):
        """Add a SLIMS record to a given SLIMS table"""
        record = self.db.add(table, data)
        logger.info(f"SLIMS Add: {table}/{record.pk()}")
        return record

    def update(self, table: SLIMS_TABLES, pk: int, data: dict):
        """Update a SLIMS record"""
        record = self.db.fetch_by_pk(table, pk)
        if record is None:
            raise ValueError(f'No data in SLIMS "{table}" table for pk "{pk}"')
        new_record = record.update(data)
        logger.info(f"SLIMS Update: {table}/{pk}")
        return new_record

    def rest_link(self, table: SLIMS_TABLES, **kwargs):
        """Construct a url link to a SLIMS table with arbitrary filters"""
        base_url = f"{self.url}/rest/{table}"
        queries = [f"?{k}={v}" for k, v in kwargs.items()]
        return base_url + "".join(queries)

    def add_model(
        self, model: SlimsBaseModelTypeVar, *args, **kwargs
    ) -> SlimsBaseModelTypeVar:
        """Given a SlimsBaseModel object, add it to SLIMS
        Args
            model (SlimsBaseModel): object to add
            *args (str): fields to include in the serialization
            **kwargs: passed to model.model_dump()

        Returns
            An instance of the same type of model, with data from
            the resulting SLIMS record
        """
        fields_to_include = set(args) or None
        fields_to_exclude = set(kwargs.get("exclude", []))
        fields_to_exclude.update(["pk", "attachments", "slims_api"])
        rtn = self.add(
            model._slims_table,
            model.model_dump(
                include=fields_to_include,
                exclude=fields_to_exclude,
                **kwargs,
                by_alias=True,
            ),
        )
        return type(model).model_validate(rtn)

    def update_model(self, model: SlimsBaseModel, *args, **kwargs):
        """Given a SlimsBaseModel object, update its (existing) SLIMS record

        Args
            model (SlimsBaseModel): object to update
            *args (str): fields to include in the serialization
            **kwargs: passed to model.model_dump()

        Returns
            An instance of the same type of model, with data from
            the resulting SLIMS record
        """
        if model.pk is None:
            raise ValueError("Cannot update model without a pk")

        fields_to_include = set(args) or None
        rtn = self.update(
            model._slims_table,
            model.pk,
            model.model_dump(
                include=fields_to_include,
                by_alias=True,
                **kwargs,
            ),
        )
        return type(model).model_validate(rtn)

    def add_attachment_content(
        self,
        record: SlimsBaseModel,
        name: str,
        content: bytes | str,
    ) -> int:
        """Add an attachment to a SLIMS record

        Returns
        -------
        int: Primary key of the attachment added.

        Notes
        -----
        - Returned attachment does not contain the name of the attachment in
         Slims, this requires a separate fetch.
        """
        if record.pk is None:
            raise ValueError("Cannot add attachment to a record without a pk")

        if isinstance(content, str):
            content = content.encode("utf-8")

        response = self.db.slims_api.post(
            url="repo",
            body={
                "attm_name": name,
                "atln_recordPk": record.pk,
                "atln_recordTable": record._slims_table,
                "contents": base64.b64encode(content).decode("utf-8"),
            },
        )
        response.raise_for_status()
        return int(response.text)



"""Contains a model for a unit"""

from datetime import datetime
from typing import Optional

from pydantic import Field

from aind_slims_api.models.base import SlimsBaseModel


class SlimsUnit(SlimsBaseModel):
    """Model for unit information in SLIMS"""

    name: str = Field(
        ...,
        serialization_alias="unit_name",
        validation_alias="unit_name",
    )
    abbreviation: Optional[str] = Field(
        "",
        serialization_alias="unit_abbreviation",
        validation_alias="unit_abbreviation",
    )
    pk: int = Field(
        ...,
        serialization_alias="unit_pk",
        validation_alias="unit_pk",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="unit_createdOn",
        validation_alias="unit_createdOn",
    )
    _slims_table = "Unit"



"""Contains a model for the mouse content, and a method for fetching it"""

from datetime import datetime
from typing import Annotated, ClassVar, Optional

from pydantic import BeforeValidator, Field

from aind_slims_api.models.base import SlimsBaseModel
from aind_slims_api.models.utils import UnitSpec


class SlimsMouseContent(SlimsBaseModel):
    """Model for an instance of the Mouse ContentType

    Properties
    ----------
    barcode: str, barcode of the mouse, filterable

    Examples
    --------
    >>> from aind_slims_api.core import SlimsClient
    >>> client = SlimsClient()
    >>> mouse = client.fetch_model(SlimsMouseContent, barcode="00000000")
    """

    baseline_weight_g: Annotated[float | None, UnitSpec("g")] = Field(
        ...,
        serialization_alias="cntn_cf_baselineWeight",
        validation_alias="cntn_cf_baselineWeight",
    )
    point_of_contact: Optional[str] = Field(
        ...,
        serialization_alias="cntn_cf_scientificPointOfContact",
        validation_alias="cntn_cf_scientificPointOfContact",
    )
    water_restricted: Annotated[bool, BeforeValidator(lambda x: x or False)] = Field(
        ...,
        serialization_alias="cntn_cf_waterRestricted",
        validation_alias="cntn_cf_waterRestricted",
    )
    barcode: str = Field(
        ...,
        serialization_alias="cntn_barCode",
        validation_alias="cntn_barCode",
    )
    pk: Optional[int] = Field(
        default=None,
        serialization_alias="cntn_pk",
        validation_alias="cntn_pk",
    )
    status: Optional[int] = Field(
        default=28,
        serialization_alias="cntn_fk_status",
        validation_alias="cntn_fk_status",
    )
    content_type: Optional[int] = Field(
        default=5,
        serialization_alias="cntn_fk_contentType",
        validation_alias="cntn_fk_contentType",
    )
    created_on: Optional[datetime] = Field(
        None,
        serialization_alias="cntn_createdOn",
        validation_alias="cntn_createdOn",
    )
    _slims_table = "Content"
    _base_fetch_filters: ClassVar[dict[str, str]] = {
        "cntp_name": "Mouse",
    }

    # TODO: Include other helpful fields (genotype, gender...)

    # pk: callable
    # cntn_fk_category: SlimsColumn
    # cntn_fk_contentType: SlimsColumn
    # cntn_barCode: SlimsColumn
    # cntn_id: SlimsColumn
    # cntn_cf_contactPerson: SlimsColumn
    # cntn_status: SlimsColumn
    # cntn_fk_status: SlimsColumn
    # cntn_fk_user: SlimsColumn
    # cntn_cf_fk_fundingCode: SlimsColumn
    # cntn_cf_genotype: SlimsColumn
    # cntn_cf_labtracksId: SlimsColumn
    # cntn_cf_parentBarcode: SlimsColumn



"""Tests methods in mouse module"""

import json
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from slims.internal import Record

from aind_slims_api.core import SlimsClient
from aind_slims_api.models.mouse import SlimsMouseContent

RESOURCES_DIR = Path(os.path.dirname(os.path.realpath(__file__))) / "resources"


class TestMouse(unittest.TestCase):
    """Tests top level methods in mouse module"""

    example_client: SlimsClient
    example_fetch_mouse_response: list[Record]

    @classmethod
    def setUpClass(cls):
        """Load json files of expected responses from slims"""
        example_client = SlimsClient(
            url="http://fake_url", username="user", password="pass"
        )
        cls.example_client = example_client
        with open(RESOURCES_DIR / "example_fetch_mouse_response.json", "r") as f:
            response = [
                Record(json_entity=r, slims_api=example_client.db.slims_api)
                for r in json.load(f)
            ]
        cls.example_fetch_mouse_response = response

    @patch("slims.slims.Slims.fetch")
    def test_fetch_mouse_content_success(self, mock_fetch: MagicMock):
        """Test fetch_mouse_content when successful"""
        mock_fetch.return_value = self.example_fetch_mouse_response
        mouse_details = self.example_client.fetch_model(
            SlimsMouseContent, barcode="123456"
        )
        self.assertEqual(
            self.example_fetch_mouse_response[0].json_entity, mouse_details.json_entity
        )


if __name__ == "__main__":
    unittest.main()



""" Library Configuration model """

from pydantic import SecretStr
from pydantic_settings import BaseSettings


class AindSlimsApiSettings(BaseSettings):
    """Settings for SLIMS Client

    Per pydantic-settings docs
    https://docs.pydantic.dev/latest/concepts/pydantic_settings/
    Loads slims credentials from environment variables if present"""

    slims_url: str = "https://aind-test.us.slims.agilent.com/slimsrest/"
    slims_username: str = ""
    slims_password: SecretStr = SecretStr("")
